class Deleted:
  def __init__(self, data: dict):
    self.id: int = data['id']
    self.channel = PsuedoChannel(data['channel_id'])
    try:
      self.guild = PsuedoGuild(data['guild_id'])
    except KeyError:
      self.guild = None

class PsuedoChannel:
  def __init__(self, id):
    self.id: int = id

class PsuedoGuild:
  def __init__(self, id):
    self.id: int = id;



import json
import time
import asyncio
import requests
import websocket
from autocord import op
from autocord import errors
from threading import Thread
from .types.Me import Me
from .types.Member import Member
from .types.Message import Message
from .types.Deleted import Deleted
from .types.DmChannel import DmChannel

class Client:
  def __init__(self, token, return_type='object'):
    self.client = requests.Session()
    self.client.headers = {'Authorization': token}
    try:
      data = self.client.get('https://discord.com/api/v9/users/@me')
      try:
        data = data.json()
        data['message']
        raise errors.UnauthorizedError(data['message'])
      except KeyError:
        self.token = token
        self.user = Me(data)

        self.on_ready = []
        self.on_message = []
        self.on_message_delete = []
        self.on_message_update = []

        self.return_type = return_type if return_type in ['object', 'json'] else 'object'
        self.ws_url = 'wss://gateway.discord.gg/?v=9&encoding=json'
        self.auth = {
          "token": self.token,
          "properties": {
              "$os": "windows",
              "$browser": "chrome",
              "$device": "pc"
          }
        }
    
    except requests.exceptions.JSONDecodeError:
      raise errors.TooManyRequests(data.text.strip())

  def event(self, func):
    if func.__name__ == 'on_ready':
      self.on_ready.append(func)
    elif func.__name__ == 'on_message':
      self.on_message.append(func)
    elif func.__name__ == 'on_message_delete':
      self.on_message_delete.append(func)
    elif func.__name__ == 'on_message_update':
      self.on_message_update.append(func)

  def send_ws(self, event, payload):
    self.ws.send(json.dumps({"op": event, "d": payload}))

  def recieve_messages(self):
    while True:
      data = json.loads(self.ws.recv())
      if data["op"] == op.DISPATCH:     
        if data['t'] == 'MESSAGE_DELETE':
          if self.return_type == 'json': data = data['d']
          else: data = Deleted(data['d'])
          for func in self.on_message_delete:
            asyncio.run(func(data))

        elif data['t'] == 'MESSAGE_UPDATE':
          if self.return_type == 'json': data = data['d']
          else: data = Message(data['d'], self)
          for func in self.on_message_update:
            asyncio.run(func(data))
        
        elif data["t"] == "MESSAGE_CREATE":
          if self.return_type == 'json': data = data['d']
          else: data = Message(data['d'], self)
          for func in self.on_message:
            asyncio.run(func(data))

  def send_heartbeat(self):
    while self.hb_interval is not None:
      self.send_ws(op.HEARTBEAT, self.auth)
      time.sleep(self.hb_interval)

  def connect(self):
    self.ws = websocket.WebSocket()
    self.ws.connect(self.ws_url)
    self.send_ws(op.IDENTIFY, self.auth)
    response = json.loads(self.ws.recv())

    if response['op'] != 10:
      raise errors.UnexpectedSocketResponse('Try running the application again')
    else:
      self.hb_interval = (response["d"]["heartbeat_interval"]-2000)/1000
      Thread(target=self.send_heartbeat).start()
      Thread(target=self.recieve_messages).start()

  # Client functions
  async def send(self, id: int, content: str=None, attachments: list=None, reference: dict=None, mention: bool=True):
    payload = {'content': content, 'allowed_mentions': {'replied_user': mention}}
    if reference:
      payload['message_reference'] = reference
      
    data = self.client.post(
      f'https://discord.com/api/v9/channels/{id}/messages',
      json=payload
    )
    data = data.json()
    try:
      data['errors']
      if 'channel_id' in data['errors']:
        raise errors.InvalidFormBody(data['errors']['channel_id']['_errors'][0]['message'])
      elif 'content' in data['errors']:
        raise errors.InvalidFormBody(data['errors']['content']['_errors'][0]['message'])
      else:
        raise errors.InvalidFormBody(data['message'])
    except KeyError:
      try:
        raise errors.InvalidFormBody(data['message'])
      except KeyError:
        if self.return_type == 'object':
          return Message(data, self)
        else:
          return data

  async def delete(self, id: int, message: int):
    data = self.client.delete(f'https://discord.com/api/v9/channels/{id}/messages/{message}')
    try:
      data = data.json()
      try:
        raise errors.InvalidFormBody(data['errors']['message_id']['_errors'][0]['message'])
      except KeyError:
        try:
          raise errors.InvalidFormBody(data['errors']['channel']['_errors'][0]['message'])
        except KeyError:
          raise errors.InvalidFormBody(data['message'])
    except requests.exceptions.JSONDecodeError: pass

  async def edit(self, id: int, message: int, content: str):
    data = self.client.patch(f'https://discord.com/api/v9/channels/{id}/messages/{message}', {'content': content})
    data = data.json()
    try:
      raise errors.InvalidFormBody(data['errors']['message_id']['_errors'][0]['message'])
    except KeyError:
      try:
        raise errors.InvalidFormBody(data['errors']['channel']['_errors'][0]['message'])
      except KeyError:
        try:
          raise errors.InvalidFormBody(data['message'])
        except KeyError:
          if self.return_type == 'object':
            return Message(data, self)
          else:
            return data

  async def fetch_user(self, id: int) -> Member:
    data = self.client.get(f'https://discord.com/api/v9/users/{id}/profile')
    try:
      data = data.json()
      data = data['user']
    except KeyError:
      try:
        raise errors.InvalidFormBody(data['errors']['user_id']['_errors'][0]['message'])
      except KeyError:
        raise errors.InvalidFormBody(data['message'])

    if self.return_type == 'object':
      return Member(data)
    else:
      return data
  
  async def fetch_channel(self, id: int, cursor: int=None) -> Message:
    if cursor is None:
      data = self.client.get(f'https://discord.com/api/v9/channels/{id}/messages')
    else:
      data = self.client.get(f'https://discord.com/api/v9/channels/{id}/messages?before={cursor}')
    try:
      data = data.json()
      try:
        raise errors.InvalidFormBody(data['channel_id']['_errors'][0]['message'])
      except KeyError:
        raise errors.InvalidFormBody(data['message'])
    except ValueError: pass
    except TypeError: pass

    if self.return_type == 'object':
      return [Message(i, self) for i in data]
    else:
      return data

  async def create_dm(self, user: int):
    data = self.client.post('https://discord.com/api/v9/users/@me/channels', json={'recipients': [user]})
    data = data.json()
    try:
      try:
        raise errors.InvalidFormBody(data['errors']['recipients']['0']['_errors'][0]['message'])
      except KeyError:
        raise errors.InvalidFormBody(data['message'])
    except KeyError:
      if self.return_type == 'object':
        return DmChannel(data, self)
      else:
        return data
  
  def run(self):
    self.connect()
    for func in self.on_ready:
      asyncio.run(func())



class Member:
  def __init__(self, data: dict):
    self.username: str = data['username']
    self.public_flags: int = data['public_flags']
    self.id: int = data['id']
    self.display_name: str = data['display_name'] if data['display_name'] != 'None' else None
    self.discriminator: int = data['discriminator']
    self.avatar: str = data['avatar']

    try:
      self.avatar_decoration: str = data['avatar_decoration'] if data['avatar_decoration'] != 'None' else None
    except KeyError:
      self.avatar_decoration: str = None

  def __str__(self):
    return f'{self.username}#{self.discriminator}'



class Me:
  def __init__(self, data: dict):
    for key in data:
      try:
        setattr(Me, key, int(data[key]))
      except ValueError:
        setattr(Me, key, data[key])
      except TypeError:
        setattr(Me, key, data[key])

    self.avatar_url = f'https://cdn.discordapp.com/avatars/{self.id}/{self.avatar}.png'

  def __str__(self):
    return f'{self.username}#{self.discriminator}'



from autocord.types.Member import Member

class DmChannel:
  def __init__(self, data: dict, client=None):
    self.__client = client
    
    self.id: int = data['id']
    self.type: int = data['type']
    self.last_mesasge_id: int = data['last_message_id']
    self.flags: int = data['flags']
    self.recipients: list[Member] = [Member(i) for i in data['recipients']]

  def send(self, content: str=None, attachments: list=None, reference: dict=None, mention: bool=True):
    # depricated due to captcha
    '''
    {
      "captcha_key":[
        "captcha-required"
      ],
      "captcha_sitekey": "x",
      "captcha_service": "hcaptcha",
      "captcha_rqdata": "x",
      "captcha_rqtoken": "x"
    }
    '''
    return self.__client.send(id=self.id, content=content, attachments=attachments, reference=reference, mention=mention)
  
  def fetch_channel(self, cursor: int=None):
    return self.__client.fetch_channel(self.id, cursor)



class Attachment:
  def __init__(self, data: dict):
    self.url: str = data['url']
    self.size: int = data['size']
    self.proxy_url: str = data['proxy_url']
    self.id: int = data['id']
    self.filename: str = data['filename']
    self.content_type: str = data['content_type']

    try:
      self.width: int = data['width']
      self.height: int = data['height']
    except KeyError:
      self.width: int = None
      self.height: int = None



class Embed:
  def __init__(self, data: dict):
    self.type: str = data['type']

    try:
      self.color: int = data['color']
    except KeyError:
      self.color: int = None
    
    try:
      self.description: str = data['description']
    except KeyError:
      self.description: str = None
      
    try:
      self.title: str = data['title']
    except KeyError:
      self.title: str = None

    try:
      self.image: PsuedoImage = PsuedoImage(data['image'])
    except KeyError:
      self.image: PsuedoImage = None

    try:
      self.url: str = data['url']
    except KeyError:
      self.url: str = None

    try:
      self.footer: PsuedoFooter = PsuedoFooter(data['footer'])
    except KeyError:
      self.footer: PsuedoFooter = None

class PsuedoFooter:
  def __init__(self, data: dict):
    self.text: str = data['text']

class PsuedoImage:
  def __init__(self, data: dict):
    self.height: int = data['height']
    self.width: int = data['width']
    self.proxy_url: str = data['proxy_url']
    self.url: str = data['url']



from setuptools import setup

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="autocord",
    version="1.5.1",
    author="walker",
    description="Discord API wrapper centered around automation",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=['autocord', 'autocord.types'],
    url="https://github.com/wa1ker38552/autocord",
    install_requires=["requests", "websocket_client"],
    python_requires=">=3.7",
    py_modules=["autocord"]
)


from autocord.types.Embed import Embed
from autocord.types.Member import Member
from autocord.types.Deleted import PsuedoChannel
from autocord.types.Attachment import Attachment

class Message:
  def __init__(self, data: dict, client=None):
    self.__client = client
    
    self.type: int = data['type']
    self.tts: bool = data['tts']
    self.timestamp: str = data['timestamp']

    try:
      self.referenced_message: Message = Message(data['referenced_message']) if data['referenced_message'] != None else None
    except KeyError:
      self.referenced_message: Message = None
      
    self.pinned: bool = data['pinned']

    try:
      self.nonce: int = data['nonce']
    except KeyError:
      self.nonce: int = None
      
    self.mentions: list[Member] = [Member(i) for i in data['mentions']]
    self.mention_roles: list[PsuedoRole] = [PsuedoRole(i) for i in data['mention_roles']]
    self.id: int = data['id']
    self.flags: int = data['flags']
    self.embeds: list[Embed] = [Embed(i) for i in data['embeds']]
    self.edited_timestamp: str = data['edited_timestamp'] if data['edited_timestamp'] != 'None' else None
    self.content: str = data['content']
    # self.components: list[Component] -> Unsupported 
    self.channel: PsuedoChannel = PsuedoChannel(data['channel_id'])
    self.author: Member = Member(data['author'])
    self.attachments: list[Attachment] = [Attachment(i) for i in data['attachments']]
    
    try:
      self.guild_id: int = data['guild_id']
    except KeyError:
      self.guild_id: int = None

  def reply(self, content: str, mention: bool=True):
    return self.__client.send(self.channel.id, content=content, reference={'channel_id': self.channel.id, 'message_id': self.id}, mention=mention)

  def delete(self):
    return self.__client.delete(self.channel.id, self.id)

  def edit(self, content: str):
    return self.__client.edit(self.channel.id, self.id, content)

class PsuedoMember:
  def __init__(self, data: dict):
    self.roles: list[PsuedoRole] = [PsuedoRole(i) for i in data['roles']]
    self.premium_since: str = data['premium_since'] if data['premium_since'] != 'None' else None
    self.pending: bool = data['pending']
    self.nick: str = data['nick'] if data['nick'] != 'None' else None
    self.mute: bool = data['mute']
    self.joined_at: str = data['joined_at']
    self.flags: int = data['flags']
    self.deaf: bool = data['deaf']
    self.communication_disabled_until: str = data['communication_disabled_until'] if data['communication_disabled_until'] != 'None' else None
    self.avatar: str = data['avatar'] if data['avatar'] != 'None' else None

class PsuedoRole:
  def __init__(self, id: int):
    self.id: int = id
